@@ -20,7 +20,8 @@ var START = 0
   , PART_DATA_START = 8
   , PART_DATA = 9
   , PART_END = 10
-  , END = 11
+  , CLOSE_BOUNDARY = 11
+  , END = 12
 
   , LF = 10
   , CR = 13
@@ -30,7 +31,8 @@ var START = 0
   , A = 97
   , Z = 122
 
-var CONTENT_TYPE_RE = /^multipart\/(form-data|related);\s*boundary=(?:"([^"]+)"|([^;]+))$/i;
+var CONTENT_TYPE_RE = /^multipart\/(?:form-data|related)(?:;|$)/i;
+var CONTENT_TYPE_PARAM_RE = /;\s*([^=]+)=(?:"([^"]+)"|([^;]+))/gi;
 var FILE_EXT_RE = /(\.[_\-a-zA-Z0-9]{0,16}).*/;
 var LAST_BOUNDARY_SUFFIX_LEN = 4; // --\r\n
 
@@ -45,10 +47,11 @@ function Form(options) {
   self.finished = false;
 
   self.autoFields = !!options.autoFields;
-  self.autoFiles = !!options.autoFields;
+  self.autoFiles = !!options.autoFiles;
 
   self.maxFields = options.maxFields || 1000;
   self.maxFieldsSize = options.maxFieldsSize || 2 * 1024 * 1024;
+  self.maxFilesSize = options.maxFilesSize || Infinity;
   self.uploadDir = options.uploadDir || os.tmpDir();
   self.encoding = options.encoding || 'utf8';
   self.hash = options.hash || false;
@@ -59,6 +62,7 @@ function Form(options) {
   self.openedFiles = [];
   self.totalFieldSize = 0;
   self.totalFieldCount = 0;
+  self.totalFileSize = 0;
   self.flushing = 0;
 
   self.backpressure = false;
@@ -76,79 +80,115 @@ function Form(options) {
 }
 
 Form.prototype.parse = function(req, cb) {
+  var called = false;
   var self = this;
+  var waitend = true;
 
-  // if the user supplies a callback, this implies autoFields and autoFiles
   if (cb) {
+    // if the user supplies a callback, this implies autoFields and autoFiles
     self.autoFields = true;
     self.autoFiles = true;
+
+    var fields = {};
+    var files = {};
+    self.on('error', function(err) {
+      if (called) return;
+
+      called = true;
+
+      // wait for req events to fire
+      process.nextTick(function() {
+        if (waitend && req.readable) {
+          // dump rest of request
+          req.resume();
+          req.once('end', function() {
+            cb(err);
+          });
+          return;
+        }
+
+        cb(err);
+      });
+    });
+    self.on('field', function(name, value) {
+      var fieldsArray = fields[name] || (fields[name] = []);
+      fieldsArray.push(value);
+    });
+    self.on('file', function(name, file) {
+      var filesArray = files[name] || (files[name] = []);
+      filesArray.push(file);
+    });
+    self.on('close', function() {
+      cb(null, fields, files);
+    });
   }
 
   self.handleError = handleError;
   self.bytesExpected = getBytesExpected(req.headers);
 
-  req.on('error', handleError);
+  req.on('end', onReqEnd);
+  req.on('error', function(err) {
+    waitend = false;
+    handleError(err);
+  });
   req.on('aborted', onReqAborted);
 
+  var state = req._readableState;
+  if (req._decoder || (state && (state.encoding || state.decoder))) {
+    // this is a binary protocol
+    // if an encoding is set, input is likely corrupted
+    validationError(new Error('request encoding must not be set'));
+    return;
+  }
+
   var contentType = req.headers['content-type'];
   if (!contentType) {
-    handleError(new Error('missing content-type header'));
+    validationError(new Error('missing content-type header'));
     return;
   }
 
-  var m = contentType.match(CONTENT_TYPE_RE);
+  var m = CONTENT_TYPE_RE.exec(contentType);
   if (!m) {
-    handleError(new Error('unrecognized content-type: ' + contentType));
+    validationError(new Error('unrecognized content-type: ' + contentType));
     return;
   }
-  var boundary = m[2] || m[3];
+
+  var boundary;
+  CONTENT_TYPE_PARAM_RE.lastIndex = m.index + m[0].length - 1;
+  while ((m = CONTENT_TYPE_PARAM_RE.exec(contentType))) {
+    if (m[1].toLowerCase() !== 'boundary') continue;
+    boundary = m[2] || m[3];
+    break;
+  }
+
+  if (!boundary) {
+    validationError(new Error('content-type missing boundary: ' + require('util').inspect(m)));
+    return;
+  }
+
   setUpParser(self, boundary);
   req.pipe(self);
 
-  if (cb) {
-    var fieldsTable = {};
-    var filesTable = {};
-    var fieldsList = [];
-    var filesList = [];
-    self.on('error', function(err) {
-      cb(err);
-    });
-    self.on('field', function(name, value) {
-      fieldsTable[name] = value;
-      fieldsList.push({name: name, value: value});
-    });
-    self.on('file', function(name, file) {
-      filesTable[name] = file;
-      filesList.push(file);
-    });
-    self.on('close', function() {
-      cb(null, fieldsTable, filesTable, fieldsList, filesList);
-    });
-  }
-
   function onReqAborted() {
+    waitend = false;
     self.emit('aborted');
     handleError(new Error("Request aborted"));
   }
 
+  function onReqEnd() {
+    waitend = false;
+  }
+
   function handleError(err) {
     var first = !self.error;
     if (first) {
       self.error = err;
       req.removeListener('aborted', onReqAborted);
-
-      // welp. 0.8 doesn't support unpipe, too bad so sad.
-      // let's drop support for 0.8 soon.
-      if (req.unpipe) {
-        req.unpipe(self);
-      }
+      req.removeListener('end', onReqEnd);
     }
 
     self.openedFiles.forEach(function(file) {
-      file.ws.destroy();
-      fs.unlink(file.path, function(err) {
-        // this is already an error condition, ignore 2nd error
-      });
+      destroyFile(self, file);
     });
     self.openedFiles = [];
 
@@ -157,9 +197,15 @@ Form.prototype.parse = function(req, cb) {
     }
   }
 
+  function validationError(err) {
+    // handle error on next tick for event listeners to attach
+    process.nextTick(handleError.bind(null, err))
+  }
 };
 
 Form.prototype._write = function(buffer, encoding, cb) {
+  if (this.error) return;
+
   var self = this
     , i = 0
     , len = buffer.length
@@ -183,7 +229,11 @@ Form.prototype._write = function(buffer, encoding, cb) {
         state = START_BOUNDARY;
         /* falls through */
       case START_BOUNDARY:
-        if (index === boundaryLength - 2) {
+        if (index === boundaryLength - 2 && c === HYPHEN) {
+          index = 1;
+          state = CLOSE_BOUNDARY;
+          break;
+        } else if (index === boundaryLength - 2) {
           if (c !== CR) return self.handleError(new Error("Expected CR Received " + c));
           index++;
           break;
@@ -288,8 +338,9 @@ Form.prototype._write = function(buffer, encoding, cb) {
             // CR = part boundary
             self.partBoundaryFlag = true;
           } else if (c === HYPHEN) {
-            // HYPHEN = end boundary
-            self.lastBoundaryFlag = true;
+            index = 1;
+            state = CLOSE_BOUNDARY;
+            break;
           } else {
             index = 0;
           }
@@ -303,14 +354,6 @@ Form.prototype._write = function(buffer, encoding, cb) {
               state = HEADER_FIELD_START;
               break;
             }
-          } else if (self.lastBoundaryFlag) {
-            if (c === HYPHEN) {
-              self.onParsePartEnd();
-              self.end();
-              state = END;
-            } else {
-              index = 0;
-            }
           } else {
             index = 0;
           }
@@ -332,6 +375,16 @@ Form.prototype._write = function(buffer, encoding, cb) {
           i--;
         }
 
+        break;
+      case CLOSE_BOUNDARY:
+        if (c !== HYPHEN) return self.handleError(new Error("Expected HYPHEN Received " + c));
+        if (index === 1) {
+          self.onParsePartEnd();
+          state = END;
+        } else if (index > 1) {
+          return self.handleError(new Error("Parser has invalid state."));
+        }
+        index++;
         break;
       case END:
         break;
@@ -433,7 +486,7 @@ Form.prototype.onParseHeadersEnd = function(offset) {
   }
 
   self.totalFieldCount += 1;
-  if (self.totalFieldCount >= self.maxFields) {
+  if (self.totalFieldCount > self.maxFields) {
     return new Error("maxFields " + self.maxFields + " exceeded.");
   }
 
@@ -446,16 +499,21 @@ Form.prototype.onParseHeadersEnd = function(offset) {
   self.destStream.filename = self.partFilename;
   self.destStream.byteOffset = self.bytesReceived + offset;
   var partContentLength = self.destStream.headers['content-length'];
-  self.destStream.byteCount = partContentLength ?
-    parseInt(partContentLength, 10) :
-    (self.bytesExpected - self.destStream.byteOffset -
-     self.boundary.length - LAST_BOUNDARY_SUFFIX_LEN);
+  self.destStream.byteCount = partContentLength ? parseInt(partContentLength, 10) :
+    self.bytesExpected ? (self.bytesExpected - self.destStream.byteOffset -
+      self.boundary.length - LAST_BOUNDARY_SUFFIX_LEN) :
+    undefined;
 
   self.emit('part', self.destStream);
   if (self.destStream.filename == null && self.autoFields) {
     handleField(self, self.destStream);
   } else if (self.destStream.filename != null && self.autoFiles) {
     handleFile(self, self.destStream);
+  } else {
+    beginFlush(self);
+    self.destStream.on('end', function(){
+      endFlush(self);
+    });
   }
 }
 
@@ -493,18 +551,31 @@ function maybeClose(self) {
   }
 }
 
+function destroyFile(self, file) {
+  if (!file.ws) return;
+  file.ws.removeAllListeners('close');
+  file.ws.on('close', function() {
+    fs.unlink(file.path, function(err) {
+      if (err && !self.error) self.handleError(err);
+    });
+  });
+  file.ws.destroy();
+}
+
 function handleFile(self, fileStream) {
-  beginFlush(self);
+  if (self.error) return;
   var file = {
     fieldName: fileStream.name,
     originalFilename: fileStream.filename,
     path: uploadPath(self.uploadDir, fileStream.filename),
     headers: fileStream.headers,
   };
+  beginFlush(self); // flush to write stream
   file.ws = fs.createWriteStream(file.path);
   self.openedFiles.push(file);
   fileStream.pipe(file.ws);
   var counter = new StreamCounter();
+  var seenBytes = 0;
   fileStream.pipe(counter);
   var hashWorkaroundStream
     , hash = null;
@@ -518,6 +589,17 @@ function handleFile(self, fileStream) {
     };
     fileStream.pipe(hashWorkaroundStream);
   }
+  counter.on('progress', function() {
+    var deltaBytes = counter.bytes - seenBytes;
+    seenBytes += deltaBytes;
+    self.totalFileSize += deltaBytes;
+    if (self.totalFileSize > self.maxFilesSize) {
+      if (hashWorkaroundStream) fileStream.unpipe(hashWorkaroundStream);
+      fileStream.unpipe(counter);
+      fileStream.unpipe(file.ws);
+      self.handleError(new Error("maxFilesSize " + self.maxFilesSize + " exceeded"));
+    }
+  });
   file.ws.on('error', function(err) {
     if (!self.error) self.handleError(err);
   });
@@ -527,6 +609,10 @@ function handleFile(self, fileStream) {
     self.emit('file', fileStream.name, file);
     endFlush(self);
   });
+  beginFlush(self); // flush from file stream
+  fileStream.on('end', function(){
+    endFlush(self);
+  });
 }
 
 function handleField(self, fieldStream) {
@@ -578,7 +664,6 @@ function setUpParser(self, boundary) {
 
   self.index = null;
   self.partBoundaryFlag = false;
-  self.lastBoundaryFlag = false;
 
   self.on('finish', function() {
     if ((self.state === HEADER_FIELD_START && self.index === 0) ||
@@ -602,7 +687,15 @@ function uploadPath(baseDir, filename) {
 
 function parseFilename(headerValue) {
   var m = headerValue.match(/\bfilename="(.*?)"($|; )/i);
-  if (!m) return;
+  if (!m) {
+    m = headerValue.match(/\bfilename\*=utf-8\'\'(.*?)($|; )/i);
+    if (m) {
+      m[1] = decodeURI(m[1]);
+    }
+    else {
+      return;
+    }
+  }
 
   var filename = m[1].substr(m[1].lastIndexOf('\\') + 1);
   filename = filename.replace(/%22/g, '"');
